/*
 * Copyright (c) 2015 J.Fojtik
 * All rights reserved.
 *
 * This file is part of the uIP TCP/IP stack
 *
 */
#include <string.h>
#include <stdlib.h>
#include "dhcpd.h"

	/* uIP includes. */
#include "uip.h"
#include "../dhcpc/dhcp_int.h"


uip_ipaddr_t DHCPD_MinIP = {0x0000,0x0000};	/**< begin of DHCPD range. */
uip_ipaddr_t DHCPD_MaxIP = {0x0000,0x0000};	/**< end of DHCPD range. */


typedef struct 
{
    struct uip_udp_conn *conn;
} DHCPdStateUDP;

static DHCPdStateUDP s;


typedef enum dhcpd_status 
{
    dhcpd_NONE = 0,
    dhcpd_OFFER,
    dhcpd_ACK
} dhcpd_status_t;


typedef struct
{
  uip_ipaddr_t  Ipaddr;		/**< IP address that can be leased. */
  u8_t		LeaseStatus;	/**< Status of DHCP daemon. */
  u8_t 		chaddr[16];	/**< Device MAC address. */
} TDhcpdTable;


TDhcpdTable *DhcpdTbl = NULL;
unsigned DhcpdTblSz = 0;


static const u8_t DHCPD_LEASE_TIME[4] = {0,0,0xA8,0xC0};


static u8_t DHCPD_parse_options(u8_t *optptr, int len)
{
  u8_t *end = optptr + len;
  u8_t type = 0;

  while(optptr < end) {
    switch(*optptr) {
    case DHCP_OPTION_MSG_TYPE:
      type = *(optptr + 2);
      break;
    case DHCP_OPTION_END:
      return type;
    }
    optptr += optptr[1] + 2;
  }
  return type;  
}


static u8_t *DHCPD_add_server_id(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_SERVER_ID;
  *optptr++ = 4;
  memcpy(optptr, uip_hostaddr, 4);
  return optptr + 4;
}


static u8_t *DHCPD_add_lease_time(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_LEASE_TIME;
  *optptr++ = 4;
  memcpy(optptr, &DHCPD_LEASE_TIME, 4);
  return optptr + 4;
}


static u8_t *DHCPD_add_subnet_mask(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_SUBNET_MASK;
  *optptr++ = 4;
  memcpy(optptr, uip_netmask, 4);
  return optptr + 4;
}


static u8_t *DHCPD_add_router(u8_t *optptr)
{
  *optptr++ = DHCP_OPTION_ROUTER;
  *optptr++ = 4;
  memcpy(optptr, uip_draddr, 4);
  return optptr + 4;
}


/** Callback from uIP stack for UDP connection. */
static void dhcpd_udp_appcall(void)
{
int i;
int SlotFound = -1;
dhcp_msg *stat;
u8_t *end;
u8_t MsgType;

  if(!uip_newdata()) return;
      
  if(uip_datalen()<=0) return;
  
  stat = (dhcp_msg*)uip_appdata;
  if(stat->op!=DHCP_REQUEST) return;
      
  MsgType = DHCPD_parse_options(&stat->options[4], uip_datalen()  - ((u8_t*)&stat->options[4]-(u8_t*)uip_appdata));
  if(MsgType!=DHCPDISCOVER && MsgType!=DHCPREQUEST && MsgType!=DHCPDECLINE) return;

  for(i=0; i<DhcpdTblSz; i++)
  {
    if(!memcmp(stat->chaddr, DhcpdTbl[i].chaddr, sizeof(DhcpdTbl[i].chaddr)))
    {
      SlotFound = i;			// This computer already asked for IP assignment.
      if(MsgType==DHCPDECLINE)
      {
        DhcpdTbl[i].LeaseStatus = dhcpd_NONE;
      }      
      break;
    }
    if(DhcpdTbl[i].LeaseStatus==dhcpd_NONE) SlotFound = i;
  }

  //#define	INFO_EVENT		0x6000
  //LogEvent(INFO_EVENT, "DHCP request %d", MsgType);

  if(SlotFound<0 || MsgType==DHCPDECLINE) return;		// No DHCP free slot.
  if(MsgType==DHCPDISCOVER)  
    DhcpdTbl[SlotFound].LeaseStatus = dhcpd_OFFER;
  else
  {
    DhcpdTbl[SlotFound].LeaseStatus = dhcpd_ACK;
  }
        
  memcpy(DhcpdTbl[SlotFound].chaddr, stat->chaddr, sizeof(DhcpdTbl[SlotFound].chaddr));
  stat->op = DHCP_REPLY;
  memcpy(&stat->yiaddr, &DhcpdTbl[SlotFound].Ipaddr, 4);	// IP address offered.        

  memcpy(stat->options, DHCP_magic_cookie, sizeof(DHCP_magic_cookie));
  
  end = DHCP_add_msg_type(&stat->options[4], (MsgType==DHCPDISCOVER)?DHCPOFFER:DHCPACK);
  end = DHCPD_add_server_id(end);
  end = DHCPD_add_lease_time(end);
  end = DHCPD_add_subnet_mask(end);
  end = DHCPD_add_router(end);
  end = DHCP_add_end(end);

  uip_udp_send(end - (u8_t*)uip_appdata);
}



/** This function initializes DHCP daemon.
 * @return	0 initialisation OK; <0 failure. */
int DHCPD_uIP_Init(void)
{
unsigned i;

  if(*(u32_t*)&DHCPD_MinIP==0 || *(u32_t*)&DHCPD_MaxIP==0) return -1;	/* No range is set. */

  if(DhcpdTbl==NULL)
  {
    DhcpdTblSz = ntohs(DHCPD_MaxIP[1]) - ntohs(DHCPD_MinIP[1]);
    if(DhcpdTblSz<=0) return -2;
    if(DhcpdTblSz>512) DhcpdTblSz = 512;
    DhcpdTbl = calloc(sizeof(TDhcpdTable),DhcpdTblSz);
    if(DhcpdTbl==NULL) return -3;
    for(i=0; i<DhcpdTblSz; i++)
    {      
      memcpy(&DhcpdTbl[i].Ipaddr, &DHCPD_MinIP, sizeof(uip_ipaddr_t));
      DhcpdTbl[i].LeaseStatus = dhcpd_NONE;
      DhcpdTbl[i].Ipaddr[1] = htons(ntohs(DhcpdTbl[i].Ipaddr[1]) + i);
    }
  }

  s.conn = uip_udp_new(NULL, 0);
  if(s.conn != NULL)
  {
    s.conn->callback = dhcpd_udp_appcall;
    uip_udp_bind(s.conn, HTONS(DHCP_SERVER_PORT)); 
  }  
  return 0;
}
